from pathlib import Path

from datasets import load_dataset
import matplotlib.pyplot as plt
import seaborn as sns
import torch
from torch.utils.data import DataLoader
import torchaudio
import numpy as np
import pytest
from scipy.stats import pearsonr
from tqdm.auto import tqdm

from vocex import Vocex
from test.wada_snr import wada_snr
from srmrpy import srmr

# def main_plot():
#     # load gigaspeech validation set in streaming mode
#     dataset = load_dataset("speechcolab/gigaspeech", "xs", use_auth_token=True, streaming=True, split="validation")

#     # load vocex model
#     vocex = Vocex.from_pretrained("models/checkpoint_half.ckpt")

#     # get the first example
#     example = next(iter(dataset))

#     # get the audio
#     audio = example["audio"]["array"]

#     # get the sample rate
#     sr = example["audio"]["sampling_rate"]

#     # run inference
#     output = vocex(audio, sr, return_activations=True, return_attention=True)

#     # plot the output
#     Path("plots").mkdir(exist_ok=True)
#     for measure in output.keys():
#         if measure not in ["dvector", "activations", "attention"]:
#             plt.figure(figsize=(20, 5))
#             mel = vocex._preprocess(audio, sr)[0]
#             plt.imshow(mel.T, aspect="auto", origin="lower", cmap="magma", alpha=0.5)
#             plt.twinx()
#             values = output[measure][0]
#             sns.lineplot(x=range(len(values)), y=values)
#             plt.title(measure)
#             plt.tight_layout()
#             plt.xlim(0, len(values))
#             plt.ylim(0, values.max()*1.5)
#             plt.savefig(f"plots/{measure}.png")
#             plt.close()

#     # test batched inference
#     batch_size = 4
#     audio_list = []
#     for item in dataset:
#         audio_list.append(item["audio"]["array"])
#         if len(audio_list) == batch_size:
#             break

#     output = vocex(audio_list, sr, return_activations=True, return_attention=True)

#     def collate_fn(batch):
#         # pad audio to the same length
#         lengths = [len(item["audio"]["array"]) for item in batch]
#         max_len = max(lengths)
#         for item in batch:
#             item["audio"]["array"] = np.pad(item["audio"]["array"], (0, max_len - len(item["audio"]["array"])))
#         return {
#             "audio": torch.tensor([item["audio"]["array"] for item in batch]),
#             "lengths": torch.tensor(lengths),
#             "sampling_rate": batch[0]["audio"]["sampling_rate"]
#         }

#     dataloader = DataLoader(dataset, batch_size=batch_size, collate_fn=collate_fn)

#     num_samples = 0
#     max_num_samples = 1000
#     for batch in dataloader:
#         output = vocex(batch["audio"], batch["sampling_rate"])
#         print(output["snr"].mean(dim=1))
#         for i, item in enumerate(batch["audio"]):
#             item = item[:batch["lengths"][i]]
#             print(wada_snr(item))
#         num_samples += batch_size
#         if num_samples >= max_num_samples:
#             break
    

# if __name__ == "__main__":
#     main_plot()

# tests
@pytest.fixture
def setup_data_and_model():
    dataset = load_dataset("speechcolab/gigaspeech", "xs", use_auth_token=True, streaming=True, split="validation")
    vocex = Vocex.from_pretrained("models/checkpoint_half.ckpt")
    def collate_fn(batch):
        # pad audio to the same length
        lengths = [len(item["audio"]["array"]) for item in batch]
        max_len = max(lengths)
        for item in batch:
            item["audio"]["array"] = np.pad(item["audio"]["array"], (0, max_len - len(item["audio"]["array"])))
        return {
            "audio": torch.tensor([item["audio"]["array"] for item in batch]),
            "lengths": torch.tensor(lengths),
            "sampling_rate": batch[0]["audio"]["sampling_rate"]
        }
    dataloader = DataLoader(dataset, batch_size=4, collate_fn=collate_fn)
    return dataset, dataloader, vocex

def test_single_inference(setup_data_and_model):
    dataset, _, vocex = setup_data_and_model
    example = next(iter(dataset))
    audio = example["audio"]["array"]
    sr = example["audio"]["sampling_rate"]
    output = vocex(audio, sr, return_activations=True, return_attention=True)
    assert output["energy"].shape == (1,712)
    assert output["snr"].shape == (1,712)
    assert output["pitch"].shape == (1,712)
    assert output["dvector"].shape == (1,256)
    assert output["activations"].shape == (1,vocex.model.hparams["measure_nlayers"],712,256)
    assert output["attention"].shape == (1,vocex.model.hparams["measure_nlayers"],712,712)

def test_single_inference_plot(setup_data_and_model):
    dataset, _, vocex = setup_data_and_model
    example = next(iter(dataset))
    audio = example["audio"]["array"]
    sr = example["audio"]["sampling_rate"]
    output = vocex(audio, sr, return_activations=True, return_attention=True)
    Path("plots").mkdir(exist_ok=True)
    for measure in output.keys():
        if measure not in ["dvector", "activations", "attention", "overall_snr", "overall_srmr"]:
            plt.figure(figsize=(20, 5))
            mel = vocex._preprocess(audio, sr)[0]
            plt.imshow(mel.T, aspect="auto", origin="lower", cmap="magma", alpha=0.5)
            plt.twinx()
            values = output[measure][0]
            sns.lineplot(x=range(len(values)), y=values)
            plt.title(measure)
            plt.tight_layout()
            plt.xlim(0, len(values))
            plt.ylim(0, values.max()*1.5)
            plt.savefig(f"plots/{measure}.png")
            plt.close()
    assert True

def test_batched_inference(setup_data_and_model):
    dataset, _, vocex = setup_data_and_model
    batch_size = 4
    audio_list = []
    sr = None
    for item in dataset:
        audio_list.append(item["audio"]["array"])
        sr = item["audio"]["sampling_rate"]
        if len(audio_list) == batch_size:
            break
    vocex.model.noise_factor = 0.0
    output = vocex(audio_list, sr, return_activations=True, return_attention=True)
    output_single = vocex(audio_list[0], sr, return_activations=True, return_attention=True)
    assert np.allclose(output["energy"][0]/output["energy"][0].max(), output_single["energy"][0]/output_single["energy"][0].max(), atol=1e-2)
    assert np.allclose(output["snr"][0]/output["snr"][0].max(), output_single["snr"][0]/output_single["snr"][0].max(), atol=1e-2)
    assert np.allclose(output["pitch"][0]/output["pitch"][0].max(), output_single["pitch"][0]/output_single["pitch"][0].max(), atol=1e-2)
    assert np.allclose(output["dvector"][0]/output["dvector"][0].max(), output_single["dvector"][0]/output_single["dvector"][0].max(), atol=1e-2)
    assert output["energy"].shape == (4,712)
    assert output["snr"].shape == (4,712)
    assert output["pitch"].shape == (4,712)
    assert output["dvector"].shape == (4,256)
    assert output["activations"].shape == (4,vocex.model.hparams["measure_nlayers"],712,256)
    assert output["attention"].shape == (4,vocex.model.hparams["measure_nlayers"],712,712)

def test_snr_correlated(setup_data_and_model):
    _, dataloader, vocex = setup_data_and_model
    num_samples = 0
    max_num_samples = 100
    snr_list = []
    wada_snr_list = []
    for batch in tqdm(dataloader, total=max_num_samples//4, desc="Testing SNR"):
        output = vocex(batch["audio"], batch["sampling_rate"])
        for i, item in enumerate(batch["audio"]):
            item = item[:batch["lengths"][i]]
            snr_list.append(output["overall_snr"][i])
            wada_snr_list.append(wada_snr(item))    
        num_samples += batch["audio"].shape[0]
        if num_samples >= max_num_samples:
            break
    plt.figure(figsize=(5,5))
    sns.scatterplot(x=snr_list, y=wada_snr_list)
    plt.xlabel("Vocex SNR")
    plt.ylabel("WADA SNR")
    plt.tight_layout()
    plt.savefig("plots/snr_correlation.png")
    corr = pearsonr(snr_list, wada_snr_list)
    print(f"SNR correlation: {corr[0]:.4f} (p={corr[1]:.4f})")
    assert corr[0] > 0.2
    assert corr[1] < 0.05

def test_srmr_correlated(setup_data_and_model):
    _, dataloader, vocex = setup_data_and_model
    num_samples = 0
    max_num_samples = 100
    srmr_list = []
    srmrpy_list = []
    for batch in tqdm(dataloader, total=max_num_samples//4, desc="Testing SRMR"):
        output = vocex(batch["audio"], batch["sampling_rate"])
        for i, item in enumerate(batch["audio"]):
            item = item[:batch["lengths"][i]]
            try:
                srmrpy_list.append(srmr(item.numpy(), batch["sampling_rate"])[0])
                srmr_list.append(output["overall_srmr"][i])
            except ValueError:
                print(f"SRMRpy failed on {i}, model output: {output['overall_srmr'][i]}")
                pass
        num_samples += batch["audio"].shape[0]
        if num_samples >= max_num_samples:
            break
    plt.figure(figsize=(5,5))
    sns.scatterplot(x=srmr_list, y=srmrpy_list)
    plt.xlabel("Vocex SRMR")
    plt.ylabel("SRMRpy SRMR")
    plt.tight_layout()
    plt.savefig("plots/srmr_correlation.png")
    corr = pearsonr(srmr_list, srmrpy_list)
    print(f"SRMR correlation: {corr[0]:.4f} (p={corr[1]:.4f})")
    assert corr[0] > 0.2
    assert corr[1] < 0.05

def test_numpy_array_input(setup_data_and_model):
    dataset, _, vocex = setup_data_and_model
    first_example = next(iter(dataset))
    assert vocex(np.random.randn(16000), 16000)["overall_snr"].shape == (1,)
    # no nan values in voice_activity_binary
    assert not np.isnan(vocex(np.random.randn(16000), 16000)["voice_activity_binary"]).any()




import numpy as np
import torch
from numba import jit
from torch.autograd import Function


@jit(nopython=True)
def compute_softdtw(D, gamma):
    B = D.shape[0]
    N = D.shape[1]
    M = D.shape[2]
    R = np.ones((B, N + 2, M + 2)) * np.inf
    R[:, 0, 0] = 0
    for k in range(B):
        for j in range(1, M + 1):
            for i in range(1, N + 1):
                r0 = -R[k, i - 1, j - 1] / gamma
                r1 = -R[k, i - 1, j] / gamma
                r2 = -R[k, i, j - 1] / gamma
                rmax = max(max(r0, r1), r2)
                rsum = np.exp(r0 - rmax) + np.exp(r1 - rmax) + np.exp(r2 - rmax)
                softmin = -gamma * (np.log(rsum) + rmax)
                R[k, i, j] = D[k, i - 1, j - 1] + softmin
    return R


@jit(nopython=True)
def compute_softdtw_backward(D_, R, gamma):
    B = D_.shape[0]
    N = D_.shape[1]
    M = D_.shape[2]
    D = np.zeros((B, N + 2, M + 2))
    E = np.zeros((B, N + 2, M + 2))
    D[:, 1 : N + 1, 1 : M + 1] = D_
    E[:, -1, -1] = 1
    R[:, :, -1] = -np.inf
    R[:, -1, :] = -np.inf
    R[:, -1, -1] = R[:, -2, -2]
    for k in range(B):
        for j in range(M, 0, -1):
            for i in range(N, 0, -1):
                a0 = (R[k, i + 1, j] - R[k, i, j] - D[k, i + 1, j]) / gamma
                b0 = (R[k, i, j + 1] - R[k, i, j] - D[k, i, j + 1]) / gamma
                c0 = (R[k, i + 1, j + 1] - R[k, i, j] - D[k, i + 1, j + 1]) / gamma
                a = np.exp(a0)
                b = np.exp(b0)
                c = np.exp(c0)
                E[k, i, j] = (
                    E[k, i + 1, j] * a + E[k, i, j + 1] * b + E[k, i + 1, j + 1] * c
                )
    return E[:, 1 : N + 1, 1 : M + 1]


class _SoftDTW(Function):
    @staticmethod
    def forward(ctx, D, gamma):
        dev = D.device
        dtype = D.dtype
        gamma = torch.Tensor([gamma]).to(dev).type(dtype)  # dtype fixed
        D_ = D.detach().cpu().numpy()
        g_ = gamma.item()
        R = torch.Tensor(compute_softdtw(D_, g_)).to(dev).type(dtype)
        ctx.save_for_backward(D, R, gamma)
        return R[:, -2, -2]

    @staticmethod
    def backward(ctx, grad_output):
        dev = grad_output.device
        dtype = grad_output.dtype
        D, R, gamma = ctx.saved_tensors
        D_ = D.detach().cpu().numpy()
        R_ = R.detach().cpu().numpy()
        g_ = gamma.item()
        E = torch.Tensor(compute_softdtw_backward(D_, R_, g_)).to(dev).type(dtype)
        return grad_output.view(-1, 1, 1).expand_as(E) * E, None


class SoftDTW(torch.nn.Module):
    def __init__(self, gamma=1.0, normalize=True):
        super(SoftDTW, self).__init__()
        self.normalize = normalize
        self.gamma = gamma
        self.func_dtw = _SoftDTW.apply

    def calc_distance_matrix(self, x, y):
        n = x.size(1)
        m = y.size(1)
        d = x.size(2)
        x = x.unsqueeze(2).expand(-1, n, m, d)
        y = y.unsqueeze(1).expand(-1, n, m, d)
        dist = torch.pow(x - y, 2).sum(3)
        return dist

    def forward(self, x, y):
        assert len(x.shape) == len(y.shape)
        squeeze = False
        if len(x.shape) < 3:
            x = x.unsqueeze(0)
            y = y.unsqueeze(0)
            squeeze = True
        if self.normalize:
            D_xy = self.calc_distance_matrix(x, y)
            out_xy = self.func_dtw(D_xy, self.gamma)
            D_xx = self.calc_distance_matrix(x, x)
            out_xx = self.func_dtw(D_xx, self.gamma)
            D_yy = self.calc_distance_matrix(y, y)
            out_yy = self.func_dtw(D_yy, self.gamma)
            result = out_xy - 1 / 2 * (out_xx + out_yy)  # distance
        else:
            D_xy = self.calc_distance_matrix(x, y)
            out_xy = self.func_dtw(D_xy, self.gamma)
            result = out_xy  # discrepancy
        return result.squeeze(0) if squeeze else result


import torch
from torch import nn
from tqdm.auto import tqdm
from .transformer import TransformerEncoder, PositionalEncoding
from .conformer_layer import ConformerLayer
from .scaler import GaussianMinMaxScaler
from .utils import Transpose, NoamLR
from .softdtw import SoftDTW

class VocexModel(nn.Module):

    def __init__(
        self,
        measures=["energy", "pitch", "srmr", "snr"],
        in_channels=80,
        filter_size=256,
        kernel_size=3,
        dropout=0.1,
        depthwise=True,
        measure_nlayers=4,
        dvector_dim=256,
        dvector_nlayers=2,
        noise_factor=0.1,
        use_softdtw=False,
        softdtw_gamma=1.0,
    ):
        super().__init__()
        self.measures = measures
        self.noise_factor = noise_factor
        in_channels = in_channels
        filter_size = filter_size
        kernel_size = kernel_size
        dropout = dropout
        depthwise = depthwise
        num_outputs = len(self.measures)

        self.loss_compounds = self.measures + ["dvector"]
        
        self.in_layer = nn.Linear(in_channels, filter_size)

        self.positional_encoding = PositionalEncoding(filter_size)

        self.layers = TransformerEncoder(
            ConformerLayer(
                filter_size,
                2,
                conv_in=filter_size,
                conv_filter_size=filter_size,
                conv_kernel=(kernel_size, 1),
                batch_first=True,
                dropout=dropout,
                conv_depthwise=depthwise,
            ),
            num_layers=measure_nlayers,
        )

        self.linear = nn.Sequential(
            nn.Linear(filter_size, 1024),
            nn.ReLU(),
            nn.Linear(1024, 1024),
            nn.ReLU(),
            nn.Linear(1024, num_outputs),
        )
        
        self.dvector_conv_in_layer = nn.Linear(filter_size, filter_size)
        self.dvector_x_in_layer = nn.Linear(in_channels, filter_size)

        self.dvector_layers = TransformerEncoder(
            ConformerLayer(
                filter_size,
                2,
                conv_in=filter_size,
                conv_filter_size=filter_size,
                conv_kernel=(kernel_size, 1),
                batch_first=True,
                dropout=dropout,
                conv_depthwise=depthwise,
            ),
            num_layers=dvector_nlayers,
        )

        dvector_input_dim = filter_size * 2
        
        self.dvector_linear = nn.Sequential(
            nn.Linear(dvector_input_dim, 1024),
            nn.ReLU(),
            nn.Linear(1024, 1024),
            nn.ReLU(),
            nn.Linear(1024, dvector_dim),
        )

        self.scaler_dict = {
            k: GaussianMinMaxScaler(10) for k in self.measures
            if not k.endswith("_binary")
        }
        self.scaler_dict["mel"] = GaussianMinMaxScaler(10)
        self.scaler_dict["dvector"] = GaussianMinMaxScaler(10)
        self.scaler_dict = nn.ModuleDict(self.scaler_dict)
        self.use_softdtw = use_softdtw
        if self.use_softdtw:
            self.softdtw = SoftDTW(gamma=softdtw_gamma)

        self.verbose = False

        self.apply(self._init_weights)

        # save hparams
        self.hparams = {
            "measures": measures,
            "in_channels": in_channels,
            "filter_size": filter_size,
            "kernel_size": kernel_size,
            "dropout": dropout,
            "depthwise": depthwise,
            "measure_nlayers": measure_nlayers,
            "dvector_dim": dvector_dim,
            "dvector_nlayers": dvector_nlayers,
            "noise_factor": noise_factor,
            "use_softdtw": use_softdtw,
            "softdtw_gamma": softdtw_gamma,
        }

    @property
    def scalers(self):
        return self.scaler_dict

    def _init_weights(self, module):
        if isinstance(module, nn.Linear):
            module.weight.data.normal_(mean=0.0, std=0.02)
            if module.bias is not None:
                module.bias.data.zero_()
        elif isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)

    def fit_scalers(self, dataloader, n_batches=1000):
        # reset scalers
        for k in self.scalers:
            self.scalers[k].is_fit = False
        for i, batch in tqdm(enumerate(dataloader), desc="Fitting scalers", total=n_batches):
            self.scalers["mel"].partial_fit(batch["mel"])
            for k in self.measures:
                if not k.endswith("_binary"):
                    self.scalers[k].partial_fit(batch["measures"][k])
            self.scalers["dvector"].partial_fit(batch["dvector"])
            if i >= n_batches:
                break
        for k in self.scalers:
            self.scalers[k].is_fit = True

    def forward(self, mel, dvector=None, measures=None, inference=False, return_activations=False, return_attention=False):
        mel_padding_mask = mel.sum(dim=-1) != 0
        mel_padding_mask = mel_padding_mask.to(torch.float32)

        if return_activations:
            self.layers.return_additional_layers = list(range(self.hparams["measure_nlayers"]))

        if not self.scalers["mel"].is_fit:
            self.scalers["mel"].partial_fit(mel)
        x = self.scalers["mel"].transform(mel)
        x = x + (torch.randn_like(x) * x.std() + x.mean()) * self.noise_factor
        out = self.in_layer(x)
        if self.verbose:
            print(out.mean(), "1")
        out = self.positional_encoding(out)
        if self.verbose:
            print(out.mean(), "2")
        res = self.layers(out, src_key_padding_mask=mel_padding_mask, need_weights=return_attention)
        if return_activations:
            activations = res["activations"]
            out_conv = res["output"]
            self.layers.return_additional_layers = None
        if return_attention:
            attention = res["attention"]
            out_conv = res["output"]
        if not return_activations and not return_attention:
            out_conv = res
        if self.verbose:
            print(out_conv.mean(), "3")
        out = self.linear(out_conv)
        if self.verbose:
            print(out.mean(), "4")
        out = out.transpose(1, 2)

        measure_results = {}
        measure_true = {}
        loss_dict = {}
        for i, measure in enumerate(self.measures):
            measure_out = out[:, i]
            if not measure.endswith("_binary") and not self.scalers[measure].is_fit:
                self.scalers[measure].partial_fit(measures[measure])
            measure_results[measure] = measure_out
        if measures is not None:
            loss_dict = {}
            for measure in self.measures:
                if not measure.endswith("_binary"):
                    measure_true[measure] = self.scalers[measure].transform(measures[measure])
                else:
                    measure_true[measure] = measures[measure]
            measures_losses = []
            for measure in self.measures:
                if measure.endswith("_binary"):
                    m_loss = nn.BCEWithLogitsLoss()(measure_results[measure]*mel_padding_mask, measure_true[measure]*mel_padding_mask)
                else:
                    if not self.use_softdtw:
                        m_loss = nn.MSELoss()(measure_results[measure]*mel_padding_mask, measure_true[measure]*mel_padding_mask)
                    else:
                        if self.verbose:
                            print(measure_results[measure], measure_true[measure])
                        m_loss = self.softdtw(
                            measure_results[measure]*mel_padding_mask,
                            measure_true[measure]*mel_padding_mask,
                        ) / 1000
                loss_dict[measure] = m_loss
                measures_losses.append(m_loss)
            loss = sum(measures_losses) / len(self.measures)
        else:
            loss = None
        ### d-vector
        # predict d-vector using global average and max pooling as input
        out_conv_dvec = self.dvector_conv_in_layer(out_conv)
        x = self.dvector_x_in_layer(x)
        out_dvec = self.dvector_layers(out_conv_dvec + x, src_key_padding_mask=mel_padding_mask)
        dvector_input = torch.cat(
            [
                torch.mean(out_dvec, dim=1),
                torch.max(out_dvec, dim=1)[0],
            ],
            dim=1,
        )
        dvector_pred = self.dvector_linear(dvector_input)
        if dvector is not None:
            if not self.scalers["dvector"].is_fit:
                self.scalers["dvector"].partial_fit(dvector)
            true_dvector = self.scalers["dvector"].transform(dvector)
            dvector_loss = nn.L1Loss()(dvector_pred, true_dvector)
            loss_dict["dvector"] = dvector_loss
            if loss is not None:
                loss += dvector_loss
                loss /= 2
            else:
                loss = dvector_loss
        if not inference:
            results = {
                "loss": loss,
                "compound_losses": loss_dict,
            }
            if return_activations:
                results["activations"] = [a.detach() for a in activations]
            if return_attention:
                results["attention"] = [a.detach() for a in attention]
            return results
        else:
            # transform back to original scale
            for measure in self.measures:
                if not measure.endswith("_binary"):
                    measure_results[measure] = self.scalers[measure].inverse_transform(
                        measure_results[measure]
                    )
                else:
                    measure_results[measure] = torch.sigmoid(measure_results[measure]).detach()
            dvector_pred = self.scalers["dvector"].inverse_transform(dvector_pred)
            results = {
                "loss": loss,
                "compound_losses": loss_dict,
                "measures": measure_results,
                "dvector": dvector_pred,
            }
            if return_activations:
                results["activations"] = [a.detach() for a in activations]
            if return_attention:
                results["attention"] = [a.detach() for a in attention]
            return results


import gzip

import torch
import torchaudio.transforms as T
from librosa.filters import mel as librosa_mel
import numpy as np

from .conformer_model import VocexModel

class Vocex():
    """ 
    This is a wrapper class for the vocex model. 
    It is used to load the model and perform inference on given audio file(s).
    """

    @staticmethod
    def _drc(x, C=1, clip_val=1e-7):
        return torch.log(torch.clamp(x, min=clip_val) * C)

    @staticmethod
    def from_pretrained(model_file, compressed=True, **postprocess_kwargs):
        """ 
        Load a pretrained model from a given .pt file.
        Also accepts huggingface model names.
        """
        if compressed:
            # decompress
            with gzip.open(model_file, "rb") as f:
                checkpoint = torch.load(f)
        else:
            checkpoint = torch.load(model_file)
        model_args = checkpoint["model_args"]
        model = VocexModel(**model_args)
        model.load_state_dict(checkpoint["state_dict"])
        return Vocex(model, **postprocess_kwargs)

    def __init__(self, model, seed=42, **postprocess_kwargs):
        self.model = model
        self.mel_spectrogram = T.Spectrogram(
            n_fft=1024,
            win_length=1024,
            hop_length=256,
            pad=0,
            window_fn=torch.hann_window,
            power=2.0,
            normalized=False,
            center=True,
            pad_mode="reflect",
            onesided=True,
        )
        mel_basis = librosa_mel(
            sr=22050,
            n_fft=1024,
            n_mels=80,
            fmin=0,
            fmax=8000,
        )
        self.mel_basis = torch.from_numpy(mel_basis).float()
        self.postprocess_kwargs = postprocess_kwargs
        for measure in self.model.measures:
            if measure not in self.postprocess_kwargs:
                self.postprocess_kwargs[measure] = {
                    "use_convolution": True,
                    "convolution_window_size": 10,
                    "convolution_window": "hann",
                    "interpolate": True,
                }
                if measure == "voice_activity":
                    self.postprocess_kwargs[measure]["normalize"] = True
                else:
                    self.postprocess_kwargs[measure]["normalize"] = False
                if measure == "pitch":
                    self.postprocess_kwargs[measure]["vad_threshold_min"] = 0.5
                    self.postprocess_kwargs[measure]["vad_threshold_max"] = 1.0
                elif measure == "snr":
                    self.postprocess_kwargs[measure]["vad_threshold_min"] = 0.4
                    self.postprocess_kwargs[measure]["vad_threshold_max"] = 1.0
                    self.postprocess_kwargs[measure]["convolution_window_size"] = 200
                elif measure == "srmr":
                    self.postprocess_kwargs[measure]["vad_threshold_min"] = 0.5
                    self.postprocess_kwargs[measure]["vad_threshold_max"] = 1.0
                    self.postprocess_kwargs[measure]["convolution_window_size"] = 200
                elif measure == "energy":
                    self.postprocess_kwargs[measure]["vad_threshold_min"] = 0.0
                    self.postprocess_kwargs[measure]["vad_threshold_max"] = 1.0
                
        # set to eval mode
        self.model.eval()
        self.seed = seed
                    

    def save_checkpoint(self, path, compressed=True):
        """ Save the model to a given path. """
        # get state dict
        state_dict = self.model.state_dict()
        # get model args
        model_args = self.model.hparams

        if compressed:
            # compress
            with gzip.open(path, "wb") as f:
                torch.save({
                    "state_dict": state_dict,
                    "model_args": model_args,
                }, f)
        else:
            torch.save({
                "state_dict": state_dict,
                "model_args": model_args,
            }, path)

    @staticmethod
    def _interpolate(x):
        if isinstance(x, torch.Tensor):
            x = x.cpu().numpy()
        def nan_helper(y):
            return np.isnan(y), lambda z: z.nonzero()[0]
        nans, y = nan_helper(x)
        if not np.all(nans):
            x[nans] = np.interp(y(nans), y(~nans), x[~nans])
        else:
            x = np.zeros_like(x)
        return torch.tensor(x)

    def _preprocess(self, audio, sr=22050):
        """ Preprocess audio. """
        if isinstance(audio, list):
            # pad to same length
            max_length = max([a.shape[-1] for a in audio])
            audio = [torch.from_numpy(a).float() if isinstance(a, np.ndarray) else a for a in audio]
            audio = [torch.nn.functional.pad(a, (0, max_length - a.shape[-1]), mode="constant", value=0) for a in audio]
            # stack
            audio = torch.stack(audio)
        # convert to tensor
        elif isinstance(audio, np.ndarray):
            audio = torch.from_numpy(audio).float()
        # resample if necessary
        if sr != 22050:
            if audio.dtype == torch.int16:
                audio = audio / 32768
            elif audio.dtype == torch.float64:
                audio = audio.to(torch.float32)
            audio = T.Resample(sr, 22050)(audio)
        if audio.ndim == 1:
            # normalize unbatched
            audio = audio / audio.abs().max()
        elif audio.ndim == 2:
            # normalize batched
            audio = audio / audio.abs().max(dim=-1, keepdim=True)[0]
        elif audio.ndim == 3:
            # normalize batched and make mono
            audio = audio.mean(dim=-1)
            audio = audio / audio.abs().max(dim=-1, keepdim=True)[0]
        # convert to spectrogram
        mel = self.mel_spectrogram(audio)
        # convert to mel spectrogram
        mel = torch.matmul(self.mel_basis, mel)
        # dynamic range compression
        mel = self._drc(mel)
        # transpose
        if mel.ndim == 2:
            # add batch dimension if necessary
            mel = mel.unsqueeze(0)
        mel = mel.transpose(1, 2)
        return mel

    def postprocess(
        self,
        measure,
        vad=None,
        use_convolution=True,
        convolution_window_size=10,
        convolution_window="hann",
        interpolate=True,
        normalize=False,
        vad_threshold_min=0.0,
        vad_threshold_max=1.0,
    ):
        """ Postprocess a given measure. """
        def _interpolate(measure, vad):
            if vad is not None and (vad_threshold_min > 0 or vad_threshold_max < 1):
                # apply voice activity detection
                measure[vad < vad_threshold_min] = torch.nan
                measure[vad > vad_threshold_max] = torch.nan
                if interpolate:
                    # interpolate nan values
                    if measure.shape[0] == 1:
                        measure[0] = Vocex._interpolate(measure[0])
                    else:
                        # for batch size > 0, we want to interpolate each batch element separately
                        measure = torch.stack([Vocex._interpolate(m) for m in measure])
                else:
                    # set missing values to 0
                    measure[torch.isnan(measure)] = 0
            return measure

        if use_convolution:
            measure = _interpolate(measure, vad)
            # apply convolution
            if convolution_window == "hann":
                window = torch.hann_window(convolution_window_size)
            elif convolution_window == "hamming":
                window = torch.hamming_window(convolution_window_size)
            elif convolution_window == "blackman":
                window = torch.blackman_window(convolution_window_size)
            else:
                raise ValueError("convolution_window must be one of 'hann', 'hamming', or 'blackman'")
            window = window / window.sum()
            window = window.unsqueeze(0).unsqueeze(0)
            # use convolution from torch.nn.functional
            # pad first (using reflection padding)
            if not isinstance(measure, torch.Tensor):
                measure = torch.tensor(measure)
            if measure.shape[-1] < convolution_window_size:
                # if measure is too short, we need to pad it with zeros on 
                measure = torch.nn.functional.pad(measure, (convolution_window_size // 2, convolution_window_size // 2), mode="constant", value=0)
                print("WARNING: measure is too short, padding with zeros")
            else:
                measure = torch.nn.functional.pad(measure, (convolution_window_size // 2, convolution_window_size // 2), mode="reflect")
            if measure.ndim == 1:
                # add batch dimension if necessary
                measure = measure.unsqueeze(0)
            if measure.shape[0] == 1:
                measure = torch.nn.functional.conv1d(measure, window, padding="same")
            else:
                # for batch size > 0, we want to apply the convolution to each batch element separately
                measure = torch.stack([torch.nn.functional.conv1d(m.unsqueeze(0), window, padding="same") for m in measure]).squeeze(1)
            # remove padding
            measure = measure[:, convolution_window_size // 2:-convolution_window_size // 2]
        if normalize:
            # normalize
            measure = measure / measure.abs().max(dim=-1, keepdim=True)[0]
        measure = _interpolate(measure, vad)
        return measure

    def __call__(self, audio, sr=22050, return_activations=False, return_attention=False):
        """ Perform inference on given audio. """
        # preprocess
        mel = self._preprocess(audio, sr)
        # forward pass
        with torch.no_grad():
            if self.seed is not None:
                torch.manual_seed(self.seed)
                np.random.seed(self.seed)
            out = self.model(mel, inference=True, return_activations=return_activations, return_attention=return_attention)

        if return_activations:
            out["activations"] = torch.stack(out["activations"]).transpose(0, 1).cpu().numpy()
        
        if return_attention:
            out["attention"] = torch.stack(out["attention"]).transpose(0, 1).cpu().numpy()

        del out["loss"]
        del out["compound_losses"]

        # convert to numpy
        for measure in out["measures"].keys():
            out[measure] = out["measures"][measure].cpu().numpy()
        # do voice activity postprocessing first
        voice_activity = None
        if "voice_activity_binary" in out:
            voice_activity_vals = out["voice_activity_binary"]
            out["voice_activity_binary"] = self.postprocess(voice_activity_vals, None, **self.postprocess_kwargs["voice_activity_binary"])
            voice_activity = out["voice_activity_binary"]
        # do other postprocessing
        for measure in out["measures"].keys():
            if measure != "voice_activity_binary":
                measure_vals = out["measures"][measure]
                out[measure] = self.postprocess(measure_vals, voice_activity, **self.postprocess_kwargs[measure])
        del out["measures"]
        if "dvector" in out:
            out["dvector"] = out["dvector"].cpu().numpy()
        for measure in out.keys():
            if isinstance(out[measure], torch.Tensor):
                out[measure] = out[measure].cpu().numpy()
        if "snr" in out and "voice_activity_binary" in out:
            out["overall_snr"] = (out["snr"] * out["voice_activity_binary"]).sum(axis=-1) / out["voice_activity_binary"].sum(axis=-1)
        if "srmr" in out and "voice_activity_binary" in out:
            va_mask = out["voice_activity_binary"] > 0.5
            out["overall_srmr"] = (out["srmr"] * va_mask).sum(axis=-1) / va_mask.sum(axis=-1)
            if np.isnan(out["overall_srmr"]).any():
                out["overall_srmr"][np.isnan(out["overall_srmr"])] = out["srmr"][np.isnan(out["overall_srmr"])].mean(axis=-1)
        return out


import numpy as np
import torch
import torch.nn as nn

class GaussianMinMaxScaler(nn.Module):
    """
    A min-max scaler that does the following, given the number of samples in the dataset N:
    1. Apply as square root to the data
    2. Apply a min-max scaler to the data ranging from the expected minimum and maximum of a Gaussian distribution with samples N
    """

    def __init__(self, width, for_tensors=True, floor=1e-6, sqrt=True, n_to_fit=10_000):
        if for_tensors:
            super().__init__()
            self.isnan = torch.isnan
        else:
            self.isnan = np.isnan
        self.sqrt = sqrt
        expected_max = width / 2
        _max = float("nan")
        _min = float("nan")
        _n = 0
        for_tensors = for_tensors
        floor = floor
        _scale = float("nan")
        if for_tensors:
            # register buffers
            self.register_buffer("max", torch.tensor(_max))
            self.register_buffer("min", torch.tensor(_min))
            self.register_buffer("_scale", torch.tensor(_scale))
            self.register_buffer("expected_max", torch.tensor(expected_max))
            self.register_buffer("floor", torch.tensor(floor))
            self.register_buffer("_n", torch.tensor(_n))
            self.for_tensors = for_tensors
        else:
            self.max = _max
            self.min = _min
            self._scale = _scale
            self.expected_max = expected_max
            self.floor = floor
            self._n = _n
            self.for_tensors = for_tensors
        self.min_set = False
        self.max_set = False
        self.n_to_fit = n_to_fit
        self.is_fit = True

    def partial_fit(self, X):
        scale_change = False
        if not self.max_set:
            self.max = X.max()
            # if self.for_tensors:
            #     self.max = self.max.detach()
            scale_change = True
            self.max_set = True
        else:
            max_candidate = X.max()
            if max_candidate > self.max:
                # if self.for_tensors:
                #     self.max = max_candidate.detach()
                #else:
                self.max = max_candidate
                scale_change = True
        if not self.min_set:
            self.min = X.min()
            # if self.for_tensors:
            #     self.min = self.min.detach()
            scale_change = True
            self.min_set = True
        else:
            min_candidate = X.min()
            if min_candidate < self.min:
                # if self.for_tensors:
                #     self.min = min_candidate.detach()
                #else:
                self.min = min_candidate
                scale_change = True
        if scale_change:
            if self.for_tensors:
                if self.sqrt:
                    self._scale = torch.sqrt(self.max - self.min + self.floor)
                else:
                    self._scale = self.max - self.min + self.floor
            else:
                if self.sqrt:
                    self._scale = np.sqrt(self.max - self.min + self.floor)
                else:
                    self._scale = self.max - self.min + self.floor
        # add numpy array size to n, regardless of shape
        self._n += len(X.flatten())
        if self._n >= self.n_to_fit:
            self.is_fit = True

    def transform(self, X):
        X = X - self.min + self.floor
        if self.for_tensors:
            X = X.clamp_(min=self.floor)
            if self.sqrt:
                X = torch.sqrt(X)
            # print if any nan values
            # if torch.isnan(X).any():
            #     print("NAN values in GaussianMinMaxScaler!")
        else:
            X = np.clip(X, a_min=self.floor, a_max=None)
            if self.sqrt:
                X = np.sqrt(X)
        X = X / self._scale
        X = X - 0.5
        X = X * self.expected_max
        return X

    def inverse_transform(self, X):
        X = X.detach()
        X = X / self.expected_max
        X = X + 0.5
        X = X * self._scale
        if self.sqrt:
            X = X ** 2
        X = X + self.min - self.floor
        return X

    @property
    def is_fitted(self):
        return self.is_fit


from setuptools import setup, find_packages

setup(
    name="vocex",
    version="0.1.6",
    description="Voice Frame-Level and Utterance-Level Attribute Extraction",
    author="Christoph Minixhofer",
    author_email="christoph.minixhofer@gmail.com",
    packages=find_packages(),
    install_requires=[
        "torchaudio>=0.8.0",
        "librosa>=0.8.0",
        "numpy>=1.19.0",
        "tqdm>=4.48.2",
    ],
    python_requires=">=3.6",
    license="MIT",
)




import copy
import math

import torch
from torch import nn

class PositionalEncoding(nn.Module):
    def __init__(self, d_model, max_len=5000, dropout=0.1):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x):
        x = x + self.pe[:, :x.size(1), :].to(x.dtype)
        return self.dropout(x)

# from https://pytorch.org/docs/1.13/_modules/torch/nn/modules/transformer.html#TransformerEncoder
class TransformerEncoder(nn.Module):
    def __init__(
        self,
        encoder_layer,
        num_layers,
        norm=None,
        return_additional_layers=None,
    ):
        super(TransformerEncoder, self).__init__()
        self.layers = nn.ModuleList([copy.deepcopy(encoder_layer) for i in range(num_layers)])
        self.num_layers = num_layers
        self.norm = norm
        self.return_additional_layers = return_additional_layers

    def forward(self, src, mask=None, src_key_padding_mask=None, condition=None, need_weights=False):
            if src_key_padding_mask is not None:
                _skpm_dtype = src_key_padding_mask.dtype
                if _skpm_dtype != torch.bool and not torch.is_floating_point(src_key_padding_mask):
                    raise AssertionError("only bool and floating types of key_padding_mask are supported")
            
            output = src
            src_key_padding_mask_for_layers = src_key_padding_mask

            output_for_return = []

            if need_weights:
                weight_list = []

            for i, mod in enumerate(self.layers):
                if condition is not None:
                    output = output + condition
                if need_weights:
                    output, weights = mod(output, src_mask=mask, src_key_padding_mask=src_key_padding_mask_for_layers, need_weights=need_weights)
                    weight_list.append(weights)
                else:
                    output = mod(output, src_mask=mask, src_key_padding_mask=src_key_padding_mask_for_layers)
                if self.return_additional_layers is not None and i in self.return_additional_layers:
                    output_for_return.append(output)

            if self.norm is not None:
                output = self.norm(output)

            if self.return_additional_layers is not None:
                if need_weights:
                    return {
                        "output": output,
                        "activations": output_for_return,
                        "attention": weight_list,
                    }
                else:
                    return {
                        "output": output,
                        "activations": output_for_return,
                    }
            else:
                if need_weights:
                    return {
                        "output": output,
                        "attention": weight_list,
                    }
                else:
                    return output